# Modules
import os           # for file operations
import time 
import asyncio      
from abc import ABC, abstractmethod # abstract classes

# Custom Modules
from trader import data

path = os.getcwd().replace('\\','/')     # path script is running from

class Algorithm(ABC):
    """ Class with variables that all algorithms have"""
    Name = ""                   # Name of algorithm
    Interval = ""               # Interval for data eg 1m, 5m, 1d, 1m
    Backtest = False            # Wether to the strategy is to back test or to run  - Default False
    API_KEY = None              # Alpha Vantage API key
    Cash = 0                    # Cash allowed for algorithm to use
    StartDate = None            # Start Date for algorithm (Some Datasources don't use)
    EndDate = None              # End Date for algorithm (Some Datasources don't use)
    Data_Source = ""            # Data Source for stock infomation (Check Data sources)
    Adjusted = False            # Wether to use Adjusted data (Some Datasources don't use)  - Default False
    Active = True               # Signal wether to run/backtest
    Time = ""                   # not set by user - used for scheduling in run method

    def set_start_date(self, startDate):
        self.StartDate = startDate
    def set_end_date(self, endDate):
        self.EndDate = endDate
    def set_cash(self, cash):
        self.Cash = cash
    def set_data_source(self, data_source):
        self.Data_Source = data_source
    def set_interval(self, interval):
        self.Interval = interval
    def set_name(self, name):
        self.Name = name
    def set_active(self, active):
        self.Active = active
    def set_adjusted(self, adjusted):
        self.Adjusted = adjusted

    # Abstract Methods
    @abstractmethod
    def init(self):
        """ Abstract method to define fields at creation """
        pass

    @abstractmethod
    def on_data(self, data):
        """ Abstract method to define what happens on new data, given tuple of data (timestrap, ohlc and volume) """
        pass

    @abstractmethod
    def stats(self):
        """ Abstract method to define what stats to return """
        pass


class StockAlgorithm(Algorithm):
    """ Abstract class to store algorithm infomation and functions for stock algorithms """
    Symbol = ""                 # Stock Ticker

    # Setter
    def set_symbol(self, symbol):
        self.Symbol = symbol


class ForexAlgorithm(Algorithm):
    """ Abstract class to store algorithm infomation and functions for forex algorithms"""
    From_Currency = ""          # From Currency Ticker
    To_Currency = ""          # To Currency Ticker

    # Setter
    def set_to_currency(self, from_currency):
        self.From_Currency = from_currency
    def set_from_currency(self, to_currency):
        self.To_Currency = to_currency



async def async_backtest(algorithm):
    """" Async method to backtest given algorithm object, gets data and passes each row to on_data method - dont call"""
    if algorithm.Active == False: return
    print(f'Getting {algorithm.Name} data, from {algorithm.Data_Source}')
    if hasattr(algorithm, 'Symbol'):                # if algorithm is stock algo
        df = await getattr(data, f'get_{algorithm.Data_Source}_stock')(algorithm)

    elif hasattr(algorithm, 'From_Currency'):       # if algorithm is forex method
        df = await getattr(data, f'get_{algorithm.Data_Source}_forex')(algorithm)
        
    start_date = df.index[0]    # start date of data   
    end_date = df.index[-1]     # end date of data

    print(f'Back Testing: {algorithm.Name}: {start_date} to, {end_date} interval: {algorithm.Interval} Data-Source: {algorithm.Data_Source}')
    # need alternative, somewhat slow
    for i in range(len(df)):
        algorithm.on_data(df.iloc[i])
    print(f'Finished: {algorithm.Name}: {start_date} to, {end_date} interval: {algorithm.Interval} Data-Source: {algorithm.Data_Source}')
    algorithm.stats()


async def create_backtest_tasks():
    """ method for backtest to create async tasks of each algorithm"""
    # async create task for each item in list
    tasks = [asyncio.ensure_future(async_backtest(algo)) for algo in backtest_algorithms]
    # wait for all tasks to complete
    await asyncio.wait(tasks)


def backtest(algorithms):
    """ Backtest method passed list of algorithms, calls on data for methods"""
    global backtest_algorithms
    backtest_algorithms = algorithms
    try:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop = asyncio.get_event_loop()
        #loop.set_debug(1)
        loop.run_until_complete(create_backtest_tasks())
    except Exception as e:
        print(e)
    finally:
        loop.close()



async def async_run(algorithm):
    """ runs given algorithm object if interval time has passed - dont call (used for run method)"""
    # get algorithms interval in minutes
    if algorithm.Interval[-1] == 'm':
        if algorithm.Interval[0] == '1' and algorithm.Interval[1] == '5':   # 15 minute interval
            algo_interval_minutes = 15
        elif algorithm.Interval[0] == '1':                                  # 1 minute interval
            algo_interval_minutes = 1
        elif algorithm.Interval[0] == '5':                                  # 5 minute interval
            algo_interval_minutes = 5
        elif algorithm.Interval[0] == '3':                                  # 30 minute interval
            algo_interval_minutes = 30
        elif algorithm.Interval[0] == '6':                                  # 60 minute interval
            algo_interval_minutes = 60
    elif algorithm.Interval[-1] == 'd':                                     # 1 day interval
        algo_interval_minutes = 1440
    elif algorithm.Interval[-1] == 'w':                                     # 1 week interval
        algo_interval_minutes = 10080
    elif algorithm.Interval[-1] == 'm':                                     # 1 month interval
        algo_interval_minutes = 43200
    elif algorithm.Interval[-1] == 'y':                                     # 1 year interval
        algo_interval_minutes = 525600
    
    # if time to run algorithm
    if minutes_ran % algo_interval_minutes == 0:
        # get appropriate data
        if hasattr(algorithm, 'Symbol'):                # if algorithm is stock algo
            df = await getattr(data, f'get_{algorithm.Data_Source}_stock')(algorithm)

        elif hasattr(algorithm, 'From_Currency'):       # if algorithm is forex method
            df = await getattr(data, f'get_{algorithm.Data_Source}_forex')(algorithm)

        algorithm.on_data(df.iloc[-1]) # call on data with last row in df
        algorithm.stats()
    await asyncio.sleep(0)


async def create_run_tasks():
    """ Create of async_run method for all run algorithms"""
    tasks = [asyncio.ensure_future(async_run(algo)) for algo in run_algorithms]
    # wait for all tasks to complete
    await asyncio.wait(tasks)


def run(Algorithms):
    """ Pass list of objects runs objects, returns nothing"""
    # currently not setup to be updatable while running
    start_time = time.time()    # for scheduling getting rest of of minute
    global run_algorithms 
    run_algorithms = [x for x in Algorithms if x.Active] 
    global minutes_ran 
    minutes_ran = 0 
    while True:
        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop = asyncio.get_event_loop()
            #loop.set_debug(1)
            loop.run_until_complete(create_run_tasks())
        except Exception as e:
            print(e)
        finally:
            loop.close()

        # wait for remainder of minute
        time.sleep(60.0 - ((time.time() - start_time) % 60.0))
        minutes_ran += 1


import yfinance as yf
import pandas as pd


async def get_yfinance_date(interval, tickers, start_date, end_date):
    """
    Gets stock data from Yahoo Finance using start and end date
    Better for getting large amounts of data as 1m data is limited to 7 day period
    :param intervel eg 1d,start and end date in yy-mm-dd format and ticker:
    :return pandas dataFrame:
    """

    return yf.download(  # or pdr.get_data_yahoo(...
        # tickers list or string as well
        tickers = tickers,

        # use "period" instead of start/end
        # valid periods: 1d,5d,1mo,3mo,6mo,1y,2y,5y,10y,ytd,max
        # (optional, default is '1mo')
        # period = "ytd",
        start = start_date,
        end = end_date,

        # fetch data by interval (including intraday if period < 60 days)
        # valid intervals: 1m,2m,5m,15m,30m,60m,90m,1h,1d,5d,1wk,1mo,3mo
        # (optional, default is '1d')
        interval = interval,

        # group by ticker (to access via data['SPY'])
        # (optional, default is 'column')
        group_by = 'ticker',

        # adjust all OHLC automatically
        # (optional, default is False)
        auto_adjust = True,

        # download pre/post regular market hours data
        # (optional, default is False)
        prepost = True,

        # use threads for mass downloading? (True/False/Integer)
        # (optional, default is True)
        threads = True,

        # proxy URL scheme use use when downloading?
        # (optional, default is None)
        proxy = None
    )




async def get_yfinance_period(interval, tickers, period):
    """
    get data from yfinance using set periods
    better for getting 1m data using 7 day periods 
    :param intervel eg 1d, period eg 5d and ticker:
    :return pandas dataFrame:
    """

    return yf.download(  # or pdr.get_data_yahoo(...
        # tickers list or string as well
        tickers = tickers,

        # use "period" instead of start/end
        # valid periods: 1d,5d,1mo,3mo,6mo,1y,2y,5y,10y,ytd,max
        # (optional, default is '1mo')
        period = period,

        # fetch data by interval (including intraday if period < 60 days)
        # valid intervals: 1m,2m,5m,15m,30m,60m,90m,1h,1d,5d,1wk,1mo,3mo
        # (optional, default is '1d')
        interval = interval,

        # group by ticker (to access via data['SPY'])
        # (optional, default is 'column')
        group_by = 'ticker',

        # adjust all OHLC automatically
        # (optional, default is False)
        auto_adjust = True,

        # download pre/post regular market hours data
        # (optional, default is False)
        prepost = True,

        # use threads for mass downloading? (True/False/Integer)
        # (optional, default is True)
        threads = True,

        # proxy URL scheme use use when downloading?
        # (optional, default is None)
        proxy = None
    )

    
async def get_yfinance_stock(algorithm):  
    """passed algorithm calls appropraite api, returns df"""
    if algorithm.Interval == '1m':
        stock_df = await get_yfinance_period(algorithm.Interval, [algorithm.Symbol], '7d')
    elif algorithm.Interval == '5m':
        stock_df = await get_yfinance_period(algorithm.Interval, [algorithm.Symbol], '7d')
    elif algorithm.Interval == '15m':
        stock_df = await get_yfinance_period(algorithm.Interval, [algorithm.Symbol], '7d')
    elif algorithm.StartDate == None or algorithm.EndDate == None:
        stock_df = await get_yfinance_period(algorithm.Interval, [algorithm.Symbol], '7d')
    elif algorithm.Interval== '30m':
        stock_df = await get_yfinance_date(algorithm.Interval, [algorithm.Symbol], algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '60m':
        stock_df = await get_yfinance_date(algorithm.Interval, [algorithm.Symbol], algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1d':
        stock_df = await get_yfinance_date(algorithm.Interval, [algorithm.Symbol], algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1w':
        stock_df = await get_yfinance_date('1wk', [algorithm.Symbol], algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1m':
        stock_df = await get_yfinance_date('1mo', [algorithm.Symbol], algorithm.StartDate, algorithm.EndDate)

    # rename column names
    stock_df.columns = ['open', 'high', 'low', 'close', 'volume']
    return stock_df



async def get_yfinance_forex(algorithm):  
    """passed ticker, interval and adjusted calls appropraite api, returns df"""
    ticker = f"{algorithm.From_Currency}{algorithm.To_Currency}=X"
    forex_df = pd.DataFrame()
    if algorithm.Interval == '1m':
        forex_df = await get_yfinance_period(algorithm.Interval, ticker, '7d')
    elif algorithm.Interval == '5m':
        forex_df = await get_yfinance_period(algorithm.Interval, ticker, '7d')
    elif algorithm.Interval == '15m':
        forex_df = await get_yfinance_period(algorithm.Interval, ticker, '7d')
    elif algorithm.StartDate == None or algorithm.EndDate == None:
        forex_df = await get_yfinance_period(algorithm.Interval, ticker, '7d')
    elif algorithm.Interval== '30m':
        forex_df = await get_yfinance_date(algorithm.Interval, ticker, algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '60m':
        forex_df = await get_yfinance_date(algorithm.Interval, ticker, algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1d':
        forex_df = await get_yfinance_date(algorithm.Interval,ticker, algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1w':
        forex_df = await get_yfinance_date('1wk', ticker, algorithm.StartDate, algorithm.EndDate)
    elif algorithm.Interval == '1m':
        forex_df = await get_yfinance_date('1mo',ticker, algorithm.StartDate, algorithm.EndDate)

    # rename column names
    forex_df.columns = ['open', 'high', 'low', 'close', 'volume']
    return forex_df


import pandas as pd   
import requests

# import asyncio
from alpha_vantage.async_support.timeseries import TimeSeries
from alpha_vantage.async_support.foreignexchange import ForeignExchange


async def get_alphav_forex(algorithm):  
    """passed algorithm calls appropraite api, returns df"""
    fx = ForeignExchange(key=algorithm.API_KEY, output_format='pandas')

    forex_df = pd.DataFrame()
    if algorithm.Interval == '1m':
        forex_df, _ = await fx.get_currency_exchange_intraday(algorithm.From_Currency, algorithm.To_Currency, interval='1min', outputsize='compact')
    elif algorithm.Interval == '5m':
        forex_df, _ = await fx.get_currency_exchange_intraday(algorithm.From_Currency, algorithm.To_Currency, interval='5min', outputsize='compact')
    elif algorithm.Interval == '15m':
        forex_df, _ = await fx.get_currency_exchange_intraday(algorithm.From_Currency, algorithm.To_Currency, interval='15min', outputsize='compact')
    elif algorithm.Interval== '30m':
        forex_df, _ = await fx.get_currency_exchange_intraday(algorithm.From_Currency, algorithm.To_Currency, interval='30min', outputsize='compact')
    elif algorithm.Interval== '60m':
        forex_df, _ = await fx.get_currency_exchange_intraday(algorithm.From_Currency, algorithm.To_Currency, interval='60min', outputsize='compact')
    elif algorithm.Adjusted:
        if algorithm.Interval == '1d':
            forex_df, _ = await fx.get_currency_exchange_daily_adjusted(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')
        elif algorithm.Interval == '1w':
            forex_df, _ = await fx.get_currency_exchange_weekly_adjusted(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')
        elif algorithm.Interval == '1M':
            forex_df, _ = await fx.get_currency_exchange_monthly_adjusted(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')
    else:
        if algorithm.Interval == '1d':
            forex_df, _ = await fx.get_currency_exchange_daily(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')
        elif algorithm.Interval == '1w':
            forex_df, _ = await fx.get_currency_exchange_weekly(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')
        elif algorithm.Interval == '1M':
            forex_df, _ = await fx.get_currency_exchange_monthly(algorithm.From_Currency, algorithm.To_Currency, outputsize='compact')

    fx.close()
    # rename timestamp column to Datetime
    forex_df.rename(columns={'date':'Datetime'}, inplace=True)
    forex_df.rename(columns={'1. open':'open'}, inplace=True)
    forex_df.rename(columns={'2. high':'high'}, inplace=True)
    forex_df.rename(columns={'3. low':'low'}, inplace=True)
    forex_df.rename(columns={'4. close':'close'}, inplace=True)
    forex_df.rename(columns={'5. volume':'volume'}, inplace=True)
    # use Datetime column as index
    # forex_df.set_index('Datetime', inplace=True)
    return forex_df



# stock get methods

async def get_alphav_stock(algorithm): 
    """passed algorithm calls appropraite api, returns df"""

    ts = TimeSeries(key=algorithm.API_KEY, output_format='pandas')
    if algorithm.Interval == '1m':
        stock_df, _ = await ts.get_intraday(symbol=algorithm.Symbol,interval='1min', outputsize='full')
    elif algorithm.Interval == '5m':
        stock_df, _ = await ts.get_intraday(symbol=algorithm.Symbol,interval='5min', outputsize='full')
    elif algorithm.Interval == '15m':
        stock_df, _ = await ts.get_intraday(symbol=algorithm.Symbol,interval='15min', outputsize='full')
    elif algorithm.Interval== '30m':
        stock_df, _ = await ts.get_intraday(symbol=algorithm.Symbol,interval='30min', outputsize='full')
    elif algorithm.Interval == '60m':
        stock_df, _ = await ts.get_intraday(symbol=algorithm.Symbol,interval='60min', outputsize='full')
    elif algorithm.Adjusted:
        if algorithm.Interval == '1d':
            stock_df, _ = await ts.get_daily_adjusted(symbol=algorithm.Symbol, outputsize='full')
        elif algorithm.Interval == '1w':
            stock_df, _ = await ts.get_weekly_adjusted(symbol=algorithm.Symbol, outputsize='full')
        elif algorithm.Interval == '1m':
            stock_df, _ = await ts.get_monthly_adjusted(symbol=algorithm.Symbol, outputsize='full')
    else:
        if algorithm.Interval == '1d':
            stock_df, _ = await ts.get_daily(symbol=algorithm.Symbol, outputsize='full')
        elif algorithm.Interval == '1w':
            stock_df, _ = await ts.get_weekly(symbol=algorithm.Symbol, outputsize='full')
        elif algorithm.Interval == '1m':
            stock_df, _ = await ts.get_monthly(symbol=algorithm.Symbol, outputsize='full')
    
    await ts.close()
            
    stock_df.rename(columns={'date':'Datetime'}, inplace=True)
    stock_df.rename(columns={'1. open':'open'}, inplace=True)
    stock_df.rename(columns={'2. high':'high'}, inplace=True)
    stock_df.rename(columns={'3. low':'low'}, inplace=True)
    stock_df.rename(columns={'4. close':'close'}, inplace=True)
    stock_df.rename(columns={'5. volume':'volume'}, inplace=True)
    # use Datetime column as index
    # stock_df.set_index('Datetime', inplace=True)
    return stock_df




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.3'
DESCRIPTION = 'Run/Backtest/Create Easy Python Trading Algorithms'
LONG_DESCRIPTION = 'A package that allows easy to write, run and backtest algorithms for stocks, forex and crypto from pre-written data sources eg. alpha vantage, yfinance'

# Setting up
setup(
    name="trader-python",
    version=VERSION,
    author="Harry Ludemann",
    author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=['pandas', 'yfinance', 'alpha_vantage', 'requests'],
    keywords=['python', 'harryludemann', 'trader', 'backtest', 'trader-python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        # "Operating System :: Unix",
        # "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)